@@ -7,9 +7,15 @@
 package bitset
 
 import (
+	"bytes"
+	"compress/gzip"
 	"encoding"
+	"encoding/base64"
+	"encoding/binary"
 	"encoding/json"
+	"errors"
 	"fmt"
+	"io"
 	"math"
 	"strconv"
 	"testing"
@@ -701,6 +707,15 @@ func TestShrink(t *testing.T) {
 		t.Error("24 should be set")
 		return
 	}
+
+	b = New(110)
+	b.Set(80)
+	b.Shrink(70)
+	for _, word := range b.set {
+		if (word != 0) {
+			t.Error("word should be 0", word)
+		}
+	}
 }
 
 func TestInsertAtWithSet(t *testing.T) {
@@ -1218,10 +1233,43 @@ func TestMarshalUnmarshalBinary(t *testing.T) {
 		t.Error("Bitsets are not equal:\n\t", a.DumpAsBits(), "\n\t", b.DumpAsBits())
 		return
 	}
+
+	aSetBit := uint(128)
+	a = New(256).Set(aSetBit)
+	aExpectedMarshaledSize := 8 /* length: uint64 */ + 4 * 8 /* set : [4]uint64 */
+	aMarshaled, err := a.MarshalBinary()
+
+	if err != nil || aExpectedMarshaledSize != len(aMarshaled) || aExpectedMarshaledSize != a.BinaryStorageSize() {
+		t.Error("MarshalBinary failed to produce expected (", aExpectedMarshaledSize , ") number of bytes")
+		return
+	}
+
+	shiftAmount := uint(72)
+	// https://github.com/bits-and-blooms/bitset/issues/114
+	for i := uint(0) ; i < shiftAmount; i++ {
+		a.DeleteAt(0)
+	}
+
+	aExpectedMarshaledSize = 8 /* length: uint64 */ + 3 * 8 /* set : [3]uint64 */
+	aMarshaled, err = a.MarshalBinary()
+	if err != nil || aExpectedMarshaledSize != len(aMarshaled) || aExpectedMarshaledSize != a.BinaryStorageSize() {
+		t.Error("MarshalBinary failed to produce expected (", aExpectedMarshaledSize , ") number of bytes")
+		return
+	}
+
+	copyBinary(t, a, b)
+
+	if b.Len() != 256 - shiftAmount || !b.Test(aSetBit - shiftAmount) {
+		t.Error("Shifted bitset is not copied correctly")
+	}
 }
 
 func TestMarshalUnmarshalBinaryByLittleEndian(t *testing.T) {
 	LittleEndian()
+	defer func() {
+		// Revert when done.
+		binaryOrder = binary.BigEndian
+	}()
 	a := New(1010).Set(10).Set(1001)
 	b := new(BitSet)
 
@@ -1270,6 +1318,33 @@ func TestMarshalUnmarshalJSON(t *testing.T) {
 	}
 }
 
+func TestMarshalUnmarshalJSONWithTrailingData(t *testing.T) {
+	a := New(1010).Set(10).Set(1001)
+	data, err := json.Marshal(a)
+	if err != nil {
+		t.Errorf(err.Error())
+		return
+	}
+
+	// appending some noise
+	data = data[:len(data)-3] // remove "
+	data = append(data, []byte(`AAAAAAAAAA"`)...)
+
+	b := new(BitSet)
+	err = json.Unmarshal(data, b)
+	if err != nil {
+		t.Errorf(err.Error())
+		return
+	}
+
+	// Bitsets must be equal after marshalling and unmarshalling
+	// Do not over-reading when unmarshalling
+	if !a.Equal(b) {
+		t.Error("Bitsets are not equal:\n\t", a.DumpAsBits(), "\n\t", b.DumpAsBits())
+		return
+	}
+}
+
 func TestMarshalUnmarshalJSONByStdEncoding(t *testing.T) {
 	Base64StdEncoding()
 	a := New(1010).Set(10).Set(1001)
@@ -1308,6 +1383,28 @@ func TestSafeSet(t *testing.T) {
 	}
 }
 
+func TestSetBitsetFrom(t *testing.T) {
+	u := []uint64{2, 3, 5, 7, 11}
+	b := new(BitSet)
+	b.SetBitsetFrom(u)
+	outType := fmt.Sprintf("%T", b)
+	expType := "*bitset.BitSet"
+	if outType != expType {
+		t.Error("Expecting type: ", expType, ", gotf:", outType)
+		return
+	}
+}
+
+func TestIssue116(t *testing.T) {
+	a := []uint64{2, 3, 5, 7, 11}
+	b := []uint64{2, 3, 5, 7, 11, 0, 1}
+	bitset1 := FromWithLength(320, a)
+	bitset2 := FromWithLength(320, b)
+	if !bitset1.Equal(bitset2) || !bitset2.Equal(bitset1) {
+		t.Error("Bitsets should be equal irrespective of the underlying capacity")
+	}
+}
+
 func TestFrom(t *testing.T) {
 	u := []uint64{2, 3, 5, 7, 11}
 	b := From(u)
@@ -1437,6 +1534,34 @@ func TestCopy(t *testing.T) {
 	}
 }
 
+func TestCopyUnaligned(t *testing.T) {
+	a := New(16)
+	a.FlipRange(0, 16)
+	b := New(1)
+	a.Copy(b)
+	if b.Count() > b.Len() {
+		t.Errorf("targets copied set count (%d) should never be larger than target's length (%d)", b.Count(), b.Len())
+	}
+	if !b.Test(0) {
+		t.Errorf("first bit should still be set in copy: %+v", b)
+	}
+
+	// Test a more complex scenario with a mix of bits set in the unaligned space to verify no bits are lost.
+	a = New(32)
+	a.Set(0).Set(3).Set(4).Set(16).Set(17).Set(29).Set(31)
+	b = New(19)
+	a.Copy(b)
+
+	const expectedCount = 5
+	if b.Count() != expectedCount {
+		t.Errorf("targets copied set count: %d, want %d", b.Count(), expectedCount)
+	}
+
+	if !(b.Test(0) && b.Test(3) && b.Test(4) && b.Test(16) && b.Test(17)) {
+		t.Errorf("expected set bits are not set: %+v", b)
+	}
+}
+
 func TestCopyFull(t *testing.T) {
 	a := New(10)
 	b := &BitSet{}
@@ -1605,3 +1730,165 @@ func TestDeleteWithBitSetInstance(t *testing.T) {
 
 	}
 }
+
+func TestWriteTo(t *testing.T) {
+	const length = 9585
+	const oneEvery = 97
+	addBuf := []byte(`12345678`)
+	bs := New(length)
+	// Add some bits
+	for i := uint(0); i < length; i += oneEvery {
+		bs = bs.Set(i)
+	}
+
+	var buf bytes.Buffer
+	n, err := bs.WriteTo(&buf)
+	if err != nil {
+		t.Fatal(err)
+	}
+	wantSz := buf.Len() // Size of the serialized data in bytes.
+	if n != int64(wantSz) {
+		t.Errorf("want write size to be %d, got %d", wantSz, n)
+	}
+	buf.Write(addBuf) // Add additional data on stream.
+
+	// Generate test input for regression tests:
+	if false {
+		gzout := bytes.NewBuffer(nil)
+		gz, err := gzip.NewWriterLevel(gzout, 9)
+		if err != nil {
+			t.Fatal(err)
+		}
+		gz.Write(buf.Bytes())
+		gz.Close()
+		t.Log("Encoded:", base64.StdEncoding.EncodeToString(gzout.Bytes()))
+	}
+
+	// Read back.
+	bs = New(length)
+	n, err = bs.ReadFrom(&buf)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if n != int64(wantSz) {
+		t.Errorf("want read size to be %d, got %d", wantSz, n)
+	}
+	// Check bits
+	for i := uint(0); i < length; i += oneEvery {
+		if !bs.Test(i) {
+			t.Errorf("bit %d was not set", i)
+		}
+	}
+
+	more, err := io.ReadAll(&buf)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if !bytes.Equal(more, addBuf) {
+		t.Fatalf("extra mismatch. got %v, want %v", more, addBuf)
+	}
+}
+
+type inCompleteRetBufReader struct {
+	returnEvery int64
+	reader      io.Reader
+	offset      int64
+}
+
+func (ir *inCompleteRetBufReader) Read(b []byte) (n int, err error) {
+	if ir.returnEvery > 0 {
+		maxRead := ir.returnEvery - (ir.offset % ir.returnEvery)
+		if len(b) > int(maxRead) {
+			b = b[:maxRead]
+		}
+	}
+	n, err = ir.reader.Read(b)
+	ir.offset += int64(n)
+	return
+}
+
+func TestReadFrom(t *testing.T) {
+	addBuf := []byte(`12345678`) // Bytes after stream
+	tests := []struct {
+		length      uint
+		oneEvery    uint
+		input       string // base64+gzipped
+		wantErr     error
+		returnEvery int64
+	}{
+		{
+			length:      9585,
+			oneEvery:    97,
+			input:       "H4sIAAAAAAAC/2IAA9VCCM3AyMDAwMSACVgYGBg4sIgLMDAwKGARd2BgYGjAFB41noDx6IAJajw64IAajw4UoMajg4ZR4/EaP5pQh1g+MDQyNjE1M7cABAAA//9W5OoOwAQAAA==",
+			returnEvery: 127,
+		},
+		{
+			length:   1337,
+			oneEvery: 42,
+			input:    "H4sIAAAAAAAC/2IAA1ZLBgYWEIPRAUQKgJkMcCZYisEBzkSSYkSTYqCxAYZGxiamZuYWgAAAAP//D0wyWbgAAAA=",
+		},
+		{
+			length:   1337, // Truncated input.
+			oneEvery: 42,
+			input:    "H4sIAAAAAAAC/2IAA9VCCM3AyMDAwARmAQIAAP//vR3xdRkAAAA=",
+			wantErr:  io.ErrUnexpectedEOF,
+		},
+		{
+			length:   1337, // Empty input.
+			oneEvery: 42,
+			input:    "H4sIAAAAAAAC/wEAAP//AAAAAAAAAAA=",
+			wantErr:  io.ErrUnexpectedEOF,
+		},
+	}
+
+	for i, test := range tests {
+		t.Run(fmt.Sprint(i), func(t *testing.T) {
+			fatalErr := func(err error) {
+				t.Helper()
+				if err != nil {
+					t.Fatal(err)
+				}
+			}
+
+			var buf bytes.Buffer
+			b, err := base64.StdEncoding.DecodeString(test.input)
+			fatalErr(err)
+			gz, err := gzip.NewReader(bytes.NewBuffer(b))
+			fatalErr(err)
+			_, err = io.Copy(&buf, gz)
+			fatalErr(err)
+			fatalErr(gz.Close())
+
+			bs := New(test.length)
+			_, err = bs.ReadFrom(&inCompleteRetBufReader{returnEvery: test.returnEvery, reader: &buf})
+			if err != nil {
+				if errors.Is(err, test.wantErr) {
+					// Correct, nothing more we can test.
+					return
+				}
+				t.Fatalf("did not get expected error %v, got %v", test.wantErr, err)
+			} else {
+				if test.wantErr != nil {
+					t.Fatalf("did not get expected error %v", test.wantErr)
+				}
+			}
+			fatalErr(err)
+
+			// Test if correct bits are set.
+			for i := uint(0); i < test.length; i++ {
+				want := i%test.oneEvery == 0
+				got := bs.Test(i)
+				if want != got {
+					t.Errorf("bit %d was %v, should be %v", i, got, want)
+				}
+			}
+
+			more, err := io.ReadAll(&buf)
+			fatalErr(err)
+
+			if !bytes.Equal(more, addBuf) {
+				t.Errorf("extra mismatch. got %v, want %v", more, addBuf)
+			}
+		})
+	}
+}
